from typing import Callable

from .rqmodel import RQModel, RQComputedModel


class RQBool(RQModel):
    """Reactive Boolean Model

    Represents a Boolean
    """

    def __init__(self, state: bool):
        """Constructor

        Args:
            state: Initial state of the model
        """
        super(RQBool, self).__init__()
        self._bool: bool = bool(state)
        """Model store variable"""

    def get(self) -> bool:
        """Get value of the model

        Returns:
            bool: value of the model

        """
        return self._bool

    def set(self, value: bool) -> None:
        """Set value of model

        Will propagate the change to the widgets linked to the model

        Args:
            value: new value of the model
        """
        self._bool = bool(value)
        self.rq_data_changed.emit()

    def toggle(self) -> None:
        """Toggle the value of the model"""
        self.set(not self.get())

    def __bool__(self) -> bool:
        """Get value of the model in bool format

        Returns:
            bool: value of the model

        """
        return self.get()

    def __str__(self) -> str:
        """Get value of the model in string format

        Returns:
            str: value of the model converted to string
        """
        return str(self.get())


class RQComputedBool(RQComputedModel, RQBool):
    """Reactive Computed Boolean Model"""

    def __init__(self, function: Callable, **kwargs):
        """Constructor

        Args:
            function: function to calculate the model value from input values

            **kwargs: reactive models in the function by variable name as keyword
                Changes in these models will trigger recalculation of the function
       """
        RQBool.__init__(self, False)
        RQComputedModel.__init__(self, function, **kwargs)

    def get(self) -> bool:
        """Get the computed value"""
        return bool(super().get())



from typing import Callable

from .rqmodel import RQModel, RQComputedModel


class RQFloat(RQModel):
    """Reactive Float Model

    Represents a floating point number
    """

    def __init__(self, initial_float: float):
        """Constructor

        Args:
            initial_float: Initial value of the model
        """
        super(RQFloat, self).__init__()
        self._float: float = float(initial_float)
        """Model store variable"""

    def get(self) -> float:
        """Get value of the model

        Returns:
            float: value of the model
        """
        return self._float

    def set(self, value: float) -> None:
        """Set value of model

        Will propagate the change to the widgets linked to the model

        Args:
            value: new value of the model
        """
        self._float = float(value)
        self.rq_data_changed.emit()

    def __str__(self) -> str:
        """Get value of the model in string format

        Returns:
            str: value of the model converted to string
        """
        return str(self.get())

    def __int__(self) -> int:
        """Get value of the model in integer format

        Returns:
            str: value of the model converted to integer
        """
        return int(self.get())

    def __float__(self) -> float:
        """
        Get value of the model in float format

        Returns:
            str: value of the model
        """
        return self.get()


class RQComputedFloat(RQComputedModel, RQFloat):
    """Reactive Computed Float Model"""

    def __init__(self, function: Callable, **kwargs):
        """Constructor

        Args:
            function: function to calculate the model value from input values

            **kwargs: reactive models in the function by variable name as keyword
                Changes in these models will trigger recalculation of the function
       """
        RQFloat.__init__(self, 0)
        RQComputedModel.__init__(self, function, **kwargs)

    def get(self) -> float:
        """Get the computed value"""
        return float(super().get())



from typing import TYPE_CHECKING, Union

import difflib

if TYPE_CHECKING:
    from .rqlist import RQList


def sequence_matching(modifiable_list: Union[list, 'RQList'], target_list: list):
    """
    Modifies a list so it's the same as an other one, applying insert and delete operations

    Args:
        modifiable_list: list that has to be modified
        target_list: list that must be copied
    """
    seq_mat = difflib.SequenceMatcher(a=modifiable_list, b=target_list)  # type: ignore

    operations = seq_mat.get_opcodes()

    running_shift = 0

    for code, i1, i2, j1, j2 in operations:

        if code == 'equal':
            # modifiable_list[i1:i2] == target_list[j1:j2] (the sub-sequences are equal).
            pass
        elif code == 'delete':
            # modifiable_list[i1:i2] should be deleted. Note that j1 == j2 in this case.

            del modifiable_list[i1 + running_shift:i2 + running_shift]
            running_shift += i1 - i2

        elif code == 'replace':
            # modifiable_list[i1:i2] should be replaced by target_list[j1:j2].

            # Delete
            del modifiable_list[i1 + running_shift:i2 + running_shift]

            # Insert
            for k, item in enumerate(target_list[j1:j2]):
                modifiable_list.insert(i1 + running_shift + k, item)

            running_shift += (j2 - j1) - (i2 - i1)

        elif code == 'insert':
            # target_list[j1:j2] should be inserted at modifiable_list[i1:i1]. Note that i1 == i2 in this case.

            for k, item in enumerate(target_list[j1:j2]):
                modifiable_list.insert(i1 + running_shift + k, item)

            running_shift += (j2 - j1)
        else:
            raise NotImplementedError(code)



from typing import TYPE_CHECKING, Union

from qtpy.QtCore import Slot  # type: ignore
from qtpy.QtWidgets import QLabel  # type: ignore

if TYPE_CHECKING:
    from PyQt5.QtCore import pyqtSlot as Slot
    from PyQt5.QtWidgets import QLabel

from .rqwidget import RQWidget
from ..models import RQModel, RQBool, RQObject


class RQLabel(RQWidget, QLabel):
    """Reactive Label Widget"""

    model: RQModel
    """Model linked to the widget"""

    def __init__(self,
                 model: Union[RQModel, str, RQObject],
                 *args,
                 rq_if: Union[RQBool, None] = None,
                 rq_disabled: Union[RQBool, None] = None,
                 **kwargs
                 ):
        """Constructor.

        Args:
            model: Model to link the widget to
            *args: arguments to pass to the native pyqt widget
            rq_if: RQBool that controls the visibility
            rq_disabled: RQBool that controls the disabling
            **kwargs: arguments to pass to the native pyqt widget
        """
        RQWidget.__init__(self, model, rq_if, rq_disabled)
        QLabel.__init__(self, str(self.model), *args, **kwargs)
        self.rq_init_widget()

        self.model.rq_data_changed.connect(self._rq_data_changed)
        self._rq_data_changed()

    @Slot()
    def _rq_data_changed(self) -> None:
        """Slot triggered when the model changes value.

        Updates value of the dial
        """
        self.setText(str(self.model))



from typing import TYPE_CHECKING, Union, Callable, Type

from qtpy.QtCore import QObject  # type: ignore
from qtpy.QtCore import Slot  # type: ignore
from qtpy.QtWidgets import QLayout, QWidget  # type: ignore

if TYPE_CHECKING:
    from PyQt5.QtCore import pyqtSlot as Slot
    from PyQt5.QtCore import QObject
    from PyQt5.QtWidgets import QLayout, QWidget

from .rqwidget import RQWidget
from ..models import RQObject


class RQWidgetObject(RQWidget, QObject):
    """Reactive Widget Object

    Displays a widget for a RQObject.
    If the instance of the RQObject changes, the widget is destroyed and a new one is created in the same place
    """

    model: RQObject
    """Model linked to the widget"""

    def __init__(self,
                 model: RQObject,
                 layout: QLayout,
                 widget: Union[Type[QWidget], Callable[[object], QWidget]]):
        """Constructor

        Args:
            model: Model to link the widget to
            layout: layout to place the widget
            widget: QWidget type that represents each the instances
                Can also be a function that accepts the instance as argument, and returns the widget instance
        """
        RQWidget.__init__(self, model)
        QObject.__init__(self)
        self.rq_init_widget()

        self.layout = layout
        if not hasattr(self.layout, "rq_widget_object"):
            self.layout.rq_widget_objects = []  # type: ignore
        self.layout.rq_widget_objects = self  # type: ignore
        if issubclass(type(widget), QWidget):
            self.rq_widget_callback = lambda instance: widget(instance)
        else:
            self.rq_widget_callback = widget
        self.model.rq_data_changed.connect(self._rq_data_changed)
        self._rq_new_widget()
        self.layout.addWidget(self.widget)

    def _rq_new_widget(self) -> QWidget:
        """Generate a new widget from the model"""
        self.widget = self.rq_widget_callback(self.model.get())
        return self.widget

    @Slot()
    def _rq_data_changed(self) -> None:
        """Slot triggered when when model has changed

        Deletes widget from the layout and adds the new one in the same place
        """
        index = self.layout.indexOf(self.widget)
        self.layout.removeWidget(self.widget)
        self.widget.deleteLater()
        self._rq_new_widget()
        self.layout.insertWidget(index, self.widget)

    def show(self):
        self.widget.show()

    def hide(self):
        self.widget.hide()



from enum import EnumMeta
from typing import Union, Iterator, Any, List

from .rqlist import RQList
from .rqmodel import RQModel


class RQChoice(RQModel):
    """Reactive Choice Model

    Represents a choice from a list of choices
    """

    def __init__(self, choices: Union[RQList, List, EnumMeta], selected: RQModel = None, allow_none=False):
        """Constructor

        Args:
            choices: Initial list of choices
            selected: Initial choice selected
        """
        super(RQChoice, self).__init__()
        self.rq_choices_list = choices
        """Reactive list of available choices"""

        self.selected: Union[Any, None] = selected
        """Selected item

        Can be None if allow_none is True
        """

        self.allow_none: bool = allow_none
        """Indicates if model accepts choice none apart from the list of choices"""

        self.validate_selected()
        if isinstance(self.rq_choices_list, RQList):
            self.rq_choices_list.rq_list_remove.connect(lambda: self.validate_selected(auto_reset=True))

    def get(self) -> Any:
        """Get current selection

        Returns:
            Any: Selected Model
        """
        return self.selected

    def get_choices(self) -> Union[RQList, List]:
        """Get list of choices

        Returns:
            List of choices
        """
        if isinstance(self.rq_choices_list, EnumMeta):
            return list(self.rq_choices_list)
        else:
            return self.rq_choices_list

    def validate_selected(self, auto_reset: bool = False) -> None:
        """Validate that the current selection is none or is a valid choice from the choices list

        Args:
            auto_reset:
                if True, when selected is not valid resets the selection
                if False, raises KeyError Exception
        """
        if self.selected is None:
            if self.allow_none:
                return
            else:
                raise ValueError

        if self.selected not in self.get_choices():
            if auto_reset:
                self.reset()
            else:
                raise KeyError

    def set(self, value: Union[Any, None]) -> None:
        """Set selected option

        Args:
            value: New selected choice
        """
        self.selected = value
        self.validate_selected()
        self.rq_data_changed.emit()

    def reset(self) -> None:
        """Reset selection to default value

        If allow_none is True default is None, else is the first element
        """
        if self.allow_none:
            self.set(None)
        else:
            self.set(self[0])

    def __str__(self) -> str:
        """Get current choice in string format

        Returns:
            str: value in string of the current choice model
        """
        return str(self.selected)

    def __iter__(self) -> Iterator[Any]:
        """Iterator of the choices of the list

        Returns:
            Iterator
        """
        return self.get_choices().__iter__()

    def __getitem__(self, index: int) -> Any:
        """Get item of the choices of the list

        Args:
            index: element of the list

        Returns:
            Any: item in the list indicated by index
        """
        return self.get_choices().__getitem__(index)



from typing import TYPE_CHECKING, Union

from qtpy.QtCore import Slot  # type: ignore
from qtpy.QtWidgets import QComboBox  # type: ignore

if TYPE_CHECKING:
    from PyQt5.QtCore import pyqtSlot as Slot
    from PyQt5.QtWidgets import QComboBox

from .rqwidget import RQWidget
from ..models import RQChoice, RQBool, RQModel, RQList


class RQComboBox(RQWidget, QComboBox):
    """Reactive ComboBox Widget"""

    model: RQChoice  # type: ignore
    """Model linked to the widget"""

    def __init__(self,
                 model: RQChoice,
                 *args,
                 rq_if: Union[RQBool, None] = None,
                 rq_disabled: Union[RQBool, None] = None,
                 **kwargs
                 ):
        """Constructor.

        Args:
            model: Model to link the widget to
            *args: arguments to pass to the native pyqt widget
            rq_if: RQBool that controls the visibility
            rq_disabled: RQBool that controls the disabling
            **kwargs: arguments to pass to the native pyqt widget
        """
        if isinstance(type(model), RQModel):
            if model.rq_read_only:  # type: ignore
                raise IOError("Cannot connect rqcombobox to a read only model")

        RQWidget.__init__(self, model, rq_if, rq_disabled)
        QComboBox.__init__(self, *args, **kwargs)
        self.rq_init_widget()

        self.model.rq_data_changed.connect(self._rq_data_changed)
        if isinstance(self.model.rq_choices_list, RQList):
            self.model.rq_choices_list.rq_list_insert.connect(self._rq_choice_insert)
            self.model.rq_choices_list.rq_list_remove.connect(self._rq_choice_remove)

        for index, choice in enumerate(self.model):
            self._rq_choice_insert(index)

        if self.model.allow_none:
            self.addItem("None")

        self.currentIndexChanged.connect(self._current_index_changed)

        self._rq_data_changed()

    @Slot(int)
    def _rq_choice_insert(self, index: int) -> None:
        """Slot triggered when the initial of choices inserts a new item.

        Adds the option to the combobox

        Args:
            index: element of the initial
        """
        self.insertItem(
            index,
            str(self.model[index])
        )

    @Slot(int)
    def _rq_choice_remove(self, index: int) -> None:
        """Slot triggered when the initial of choices removes a new item

        Removes the option from the combobox

        Args:
            index: element of the initial
        """
        self.removeItem(index)

    @Slot()
    def _rq_data_changed(self) -> None:
        """Slot triggered when the selection of the choice model changed

        Updates the combobox selection
        """
        if not self._rq_writing:
            self._rq_reading = True
            if self.model.selected is not None:
                self.setCurrentIndex(
                    self.model.get_choices().index(
                        self.model.selected
                    )
                )
            else:
                self.setCurrentIndex(
                    self.count() - 1
                )
            self._rq_reading = False

    _rq_writing = False
    """Flag to signal that this widget is triggering the update and is writing to the model"""

    _rq_reading = False
    """Flag to indicate that the model changed and the widget is reading the model"""

    @Slot(int)
    def _current_index_changed(self, index: int) -> None:
        """Slot triggered when the user changes the selection in the combobox

        Args:
            index: selected item index
        """
        if not self._rq_reading:
            self._rq_writing = True

            if self.model.allow_none and self.currentIndex() == self.count() - 1:
                choice = None
            else:
                choice = self.model[index]

            self.model.set(choice)

            self._rq_writing = False



from typing import TYPE_CHECKING, Union

from qtpy.QtCore import Slot  # type: ignore
from qtpy.QtWidgets import QCheckBox  # type: ignore

if TYPE_CHECKING:
    from PyQt5.QtCore import pyqtSlot as Slot
    from PyQt5.QtWidgets import QCheckBox

from .rqwidget import RQWidget
from ..models import RQBool, RQModel


class RQCheckBox(RQWidget, QCheckBox):
    """Reactive Checkbox Widget"""

    model: RQBool
    """Model linked to the widget"""

    def __init__(self,
                 model: Union[RQBool, bool],
                 *args,
                 rq_if: Union[RQBool, None] = None,
                 rq_disabled: Union[RQBool, None] = None,
                 **kwargs
                 ):
        """Constructor.

        Args:
            model: Model to link the widget to
            *args: arguments to pass to the native pyqt widget
            rq_if: RQBool that controls the visibility
            **kwargs: arguments to pass to the native pyqt widget
        """
        if isinstance(type(model), RQModel):
            if model.rq_read_only:  # type: ignore
                raise IOError("Cannot connect rqcheckbox to a read only model")

        RQWidget.__init__(self, model, rq_if, rq_disabled)
        QCheckBox.__init__(self, *args, **kwargs)
        self.rq_init_widget()

        self.toggled.connect(self._toggled)
        self.model.rq_data_changed.connect(self._rq_data_changed)
        self._rq_data_changed()

    @Slot()
    def _rq_data_changed(self) -> None:
        """Slot triggered when the model changes state.

        Updates the state of the checkbox
        """
        if not self._rq_writing:
            self._rq_reading = True
            self.setChecked(bool(self.model))
            self._rq_reading = False

    _rq_writing = False
    """Flag to signal that this widget is triggering the update and is writing to the model"""

    _rq_reading = False
    """Flag to indicate that the model changed and the widget is reading the model"""

    @Slot()
    def _toggled(self) -> None:
        """Slot triggered when the user changes state of this checkbox.

        Propagates changes to the model
        """
        if not self._rq_reading:
            self._rq_writing = True
            self.model.set(self.isChecked())
            self._rq_writing = False



from .models.rqobject import RQObject


def new__setattr__(self, key, value) -> None:
    """Set Attribute Replacement

    Captures changes in the attributes and reports them to the rq_reactive_attributes RQObject models

    Args:
        self: object to capture
        key: key name of the attribute
        value: new value of the attribute
    """
    super(type(self), self).__setattr__(key, value)
    if hasattr(self, "rq_reactive_attributes"):
        if self.rq_reactive_attributes is not None:
            if key in self.rq_reactive_attributes:
                self.rq_reactive_attributes[key].set(value)


def reactivize(obj_type: type) -> None:
    """Reactivize method

    Injects the set attribute replacement to a object type and creates a placeholder for the reactive attributes

    Args:
        obj_type: object type to reactivize
    """
    obj_type.__setattr__ = new__setattr__  # type: ignore
    obj_type.rq_reactive_attributes = None  # type: ignore


def rq_getattr(obj: object, attribute_name: str) -> RQObject:
    """Reactive Get Attribute

    Returns an RQObject that is linked to a object attribute. Changes are two-way propagated

    Args:
        obj: object to get the attribute from
        attribute_name: name of the attribute

    Returns:
        RQObject: reactive object
    """
    if not hasattr(obj, "rq_reactive_attributes"):
        reactivize(type(obj))
    if obj.rq_reactive_attributes is None:  # type: ignore
        obj.rq_reactive_attributes = {}  # type: ignore

    # Check if rqobject has already been created for this attribute
    if attribute_name in obj.rq_reactive_attributes:  # type: ignore
        # Return if already exists
        return obj.rq_reactive_attributes[attribute_name]  # type: ignore
    else:
        # Create the RQObject with the initial value from the attribute
        reactive_attribute = RQObject(obj.__getattribute__(attribute_name))

        # Propagate changes from the rqobject to the attribute.
        # Using the __setattr__ from the super(), to avoid new__setattr__ and re-triggering the update
        reactive_attribute.rq_data_changed.connect(
            lambda: super(type(obj), obj).__setattr__(attribute_name, reactive_attribute.get())  # type: ignore
        )

        # Store the object to the dictionary
        obj.rq_reactive_attributes[attribute_name] = reactive_attribute  # type: ignore
        return reactive_attribute



from typing import TYPE_CHECKING, Union

from qtpy.QtWidgets import QWidget  # type: ignore

if TYPE_CHECKING:
    from PyQt5.QtWidgets import QWidget

from ..models import RQModel, RQText, RQBool, RQInt, RQFloat, RQObject


class RQWidget:
    """Reactive Widget Base class"""

    def __init__(self,
                 model: Union[RQModel, str, bool, int, float, RQObject, None],
                 rq_if: Union[RQBool, None] = None,
                 rq_disabled: Union[RQBool, None] = None
                 ):
        """Constructor.

        This class always always inherits from QWidget but is not declared as so. User defined widgets,
        will directly have the QWidget methods

        Args:
            model: model to link the widget to. Can also be a built-in type that is converted to a dummy model
            rq_if: RQBool that controls the visibility
            rq_disabled: RQBool that controls the disabling
        """

        if isinstance(model, str):
            model = RQText(model)
        elif isinstance(model, int):
            model = RQInt(model)
        elif isinstance(model, float):
            model = RQFloat(model)
        elif isinstance(model, bool):
            model = RQBool(model)

        self.model = model

        self._rq_if_model = rq_if
        self._rq_disabled_model = rq_disabled

    def rq_init_widget(self):
        if self._rq_if_model is not None:
            self._rq_if_model.rq_data_changed.connect(self._rq_if_data_changed)
            self._rq_if_data_changed()
        if self._rq_disabled_model is not None:
            self._rq_disabled_model.rq_data_changed.connect(self._rq_disabled_data_changed)
            self._rq_disabled_data_changed()

    def _rq_if_data_changed(self) -> None:
        if issubclass(type(self), QWidget):
            if self._rq_if_model:
                self.show()  # type: ignore
            else:
                self.hide()  # type: ignore

    def _rq_disabled_data_changed(self) -> None:
        if issubclass(type(self), QWidget):
            self.setDisabled(bool(self._rq_disabled_model))  # type: ignore



from typing import TYPE_CHECKING, List, Iterator, Callable, Any, Dict

from qtpy.QtCore import Signal  # type: ignore

if TYPE_CHECKING:
    from PyQt5.QtCore import pyqtSignal as Signal

from .rqint import RQInt
from .rqmodel import RQModel, RQComputedModel

from .sequence_matching import sequence_matching

import weakref


class RQList(RQModel):
    """Reactive List Model

    Represents a list of model instances

    This model is quite different from the others and doesn't use data_changed signal.
    Instead it uses insert and remove signals that indicate the index of the list where this event is happening.
    This can greatly improve the efficiency of widgets that use this model.
    """

    rq_list_insert = Signal(int)
    """List insert signal. 

    Indicates that there's been an insertion to the position indicated by the int
    """

    rq_list_remove = Signal(int)
    """List remove signal. 

    Indicates that there's been an deletion in the position indicated by the int
    """

    def __init__(self, initial_items: List[Any] = None):
        """Constructor

        Args:
            initial_items: List of items
        """
        self._list: List[Any] = initial_items if initial_items is not None else []
        """Model store variable

        Stores list of instances
        """

        RQModel.__init__(self)

        self._reactive_indexes: Dict[RQModel, RQInt] = weakref.WeakKeyDictionary()  # type: ignore
        """
        Weak reference dictionary of reactive indexes requested and that
        have to be updated when list changes
        
        Key is model
        Value is reactive index
        """

        self.rq_data_changed.connect(self.update_reactive_indexes)

    def set(self, items: List[Any]) -> None:
        self.clear()
        for item in items:
            self.append(item)

    def get(self) -> list:
        """Get value of the model

        Returns:
            list: value of the model
        """
        return self._list

    def insert(self, index: int, item: Any) -> None:
        """Insert a item to the specified index on the list

        Args:
            index: positional index on the list
            item: item to be inserted

        Returns:

        """
        if isinstance(item, RQModel):
            item._rq_delete.connect(
                lambda: self.remove_all(item)
            )
        self._list.insert(index, item)
        self.rq_list_insert.emit(index)
        self.rq_data_changed.emit()

    def append(self, item: Any) -> None:
        """Append a item to the end of the list

        Args:
            item: Item to be appended
        """
        RQList.insert(
            self,
            index=len(self._list),
            item=item
        )

    def __delitem__(self, key) -> None:
        """Delete the item in the list in the specified index

        Args:
            key: positional index on the list
        """
        self._list.__delitem__(key)
        self.rq_list_remove.emit(key)
        self.rq_data_changed.emit()

    def pop(self) -> None:
        """Delete last instance of the list"""
        RQList.__delitem__(self, len(self._list) - 1)

    def remove(self, item: Any) -> None:
        """Remove first occurrence of value

        Args:
            item: item
        """
        index = RQList.index(self, item)
        self.__delitem__(index)

    def remove_all(self, item: Any) -> None:
        """Remove all instances in the list

        Args:
            item: item
        """
        while True:
            try:
                self.remove(item)
            except ValueError:
                break

    def clear(self) -> None:
        """Clear all items of the list"""
        while len(self) > 0:
            self.pop()

    def __getitem__(self, index: int) -> Any:
        """Returns the indicated item of the list

        Args:
            index: element of the list

        Returns:
            Any: item in the list indicated by index
        """
        return self._list[index]

    def index(self, item: Any) -> int:
        """Returns the index where a item is located

        Raises an ValueError if is not in the list

        Args:
            item: instance that should be in the list

        Returns:
            int: index of the item in the list
        """
        return self._list.index(item)

    def update_reactive_indexes(self):
        for model, reactive_index in self._reactive_indexes.items():
            reactive_index: RQInt
            if model in self._list:
                reactive_index.set(
                    self.index(model)
                )

    def reactive_index(self, model: Any) -> RQInt:
        """Returns a reactive index model that indicates where the item is located

        Args:
            item: instance that should be in the list

        Returns:
            RQListIndex: reactive index of the item in the list
        """
        reactive_index = RQInt(0)  # TODO
        self._reactive_indexes[model] = reactive_index
        return reactive_index

    def __iter__(self) -> Iterator[Any]:
        """Iterator of the elements of the list

        Returns:
            Iterator of RQModels
        """
        for item in self._list:
            yield item

    def __len__(self) -> int:
        """Length of the list"""
        return len(self._list)

    def count(self, value) -> int:
        """Same as python list method"""
        return self._list.count(value)

    def extend(self, iterable: List[Any]):
        """Same as python list method"""
        for item in iterable:
            self.append(item)

    def __contains__(self, item: Any) -> bool:
        """Same as python list method"""
        return self._list.__contains__(item)


class RQComputedList(RQComputedModel, RQList):
    """Reactive Computed List Model"""

    def __init__(self, function: Callable, **kwargs):
        """Constructor

        Args:
            function: function to calculate the model value from input values

            **kwargs: reactive models in the function by variable name as keyword
                Changes in these models will trigger recalculation of the function
        """
        RQList.__init__(self, [])
        RQComputedModel.__init__(self, function, **kwargs)

    def _variable_changed(self) -> None:
        """Variable changed slot

        Called when some of the models have emitted rq_data_changed.

        Recalculates the list with the function and calculates differences with the current list,
        inserting and deleting the differences.
        """
        # Recompute list
        new_list = RQComputedModel.get(self)

        # Apply operations to current list so it's the same as the newly computed
        sequence_matching(
            modifiable_list=self,
            target_list=new_list
        )

    def get(self) -> list:
        """See overridden method

        ComputedLists are recalculated on change, not on request,
        so this just redirects to the actual list.
        """
        return self._list



from pyreaqtive import RQList


def rq_getlist(obj: object, attribute_name: str) -> RQList:
    """Reactive Get List

    Replaces a list of an object by an RQList and returns it

    Warnings:
        This method replaces the list with an RQList, all methods of which are compatible
        with built-in lists. However, the signature of the object is changed from list to RQList.
        Using this method on objects that rely on isinstance or issubclass on the list, will break.

    Args:
        obj: object to the list
        attribute_name: attribute of the object that is the list

    Returns:
        RQList: reactive list object
    """
    lst = getattr(obj, attribute_name)
    if isinstance(lst, RQList):
        # Already been reactivized
        return lst
    elif not isinstance(lst, list):
        raise TypeError(f"{attribute_name} of obj is not a list")
    else:
        rq_list = RQList(lst)
        setattr(obj, attribute_name, rq_list)
        return rq_list



from abc import ABC, abstractmethod
from typing import Any, Callable, Optional

from .models import RQModel


class AbstractConversion(ABC):
    """Base model for bidirectional conversion of two models"""

    @abstractmethod
    def convert_a_to_b(self, a: Any) -> Any:
        raise NotImplemented

    @abstractmethod
    def convert_b_to_a(self, b: Any) -> Any:
        raise NotImplemented


class Conversion(AbstractConversion):
    """Conversion Declared with two functions"""

    def __init__(self,
                 a_to_b: Callable[[Any], Any],
                 b_to_a: Callable[[Any], Any]
                 ):
        self.a_to_b = a_to_b
        self.b_to_a = b_to_a

    def convert_a_to_b(self, a: Any) -> Any:
        return self.a_to_b(a)

    def convert_b_to_a(self, b: Any) -> Any:
        return self.b_to_a(b)


class LinearConversion(AbstractConversion):
    """Linear Conversion Declared with scale and offset"""

    def __init__(self, scale_a_to_b: float, offset_a_to_b: float = 0):
        self.scale_a_to_b = scale_a_to_b
        self.offset_a_to_b = offset_a_to_b

    def convert_a_to_b(self, a: Any) -> Any:
        return a * self.scale_a_to_b + self.offset_a_to_b

    def convert_b_to_a(self, b: Any) -> Any:
        return (b - self.offset_a_to_b) / self.scale_a_to_b


class RQConnect:
    """Bidirectional connection between two models"""

    def __init__(
            self,
            model_a: RQModel,
            model_b: RQModel,
            conversion: AbstractConversion
    ):
        self.model_a = model_a
        self.model_b = model_b
        self.conversion = conversion

        self._propagating = False
        self.model_a.rq_data_changed.connect(self._propagate_a_to_b)
        self.model_b.rq_data_changed.connect(self._propagate_b_to_a)
        self._propagate_a_to_b()

    def _propagate_a_to_b(self):
        if not self._propagating:
            self._propagating = True
            self.model_b.set(
                self.conversion.convert_a_to_b(
                    self.model_a.get()
                )
            )
            self._propagating = False

    def _propagate_b_to_a(self):
        if not self._propagating:
            self._propagating = True
            self.model_a.set(
                self.conversion.convert_b_to_a(
                    self.model_b.get()
                )
            )
            self._propagating = False



__version__ = "1.4.0"

from .models import \
    RQModel, RQComputedModel, \
    RQInt, RQComputedInt, \
    RQBool, RQComputedBool, \
    RQText, RQComputedText, \
    RQList, RQComputedList, \
    RQChoice, \
    RQFloat, RQComputedFloat, \
    RQObject

from .widgets import \
    RQWidget, \
    RQLabel, \
    RQSpinBox, \
    RQCheckBox, \
    RQLineEdit, \
    RQComboBox, \
    RQProgressBar, \
    RQSlider, \
    RQDial, \
    RQDoubleSpinBox, \
    RQWidgetObject, \
    RQPushButton

# Backwards compatibility
RQCheckbox = RQCheckBox
RQCombobox = RQComboBox

from .layouts import \
    RQBoxLayout, \
    RQVBoxLayout, \
    RQHBoxLayout

from .rq_getattr import rq_getattr
from .rq_getlist import rq_getlist
from .rq_connect import RQConnect
from .rq_async import RQAsync



from typing import Union, Callable

from .rqmodel import RQModel, RQComputedModel


class RQText(RQModel):
    """Reactive Text Model

    Represents a string of text
    """

    def __init__(self, text: str):
        """Constructor

        Args:
            text: Initial value of the model
        """
        super(RQText, self).__init__()
        self._text: str = text
        """Model store variable"""

    def get(self) -> str:
        """Get value of the model

        Returns:
            str: value of the model
        """
        return self._text

    def set(self, value: str) -> None:
        """Will propagate the change to the widgets linked to the model

        Args:
            value: new value of the model
        """
        self._text = value
        self.rq_data_changed.emit()

    def __str__(self) -> str:
        """Get value of the model in string format

        Returns:
            str: value of the model
        """
        return self.get()


class RQComputedText(RQComputedModel, RQText):
    """Reactive Computed Float Model"""

    def __init__(self, function: Union[str, Callable[..., str]], **kwargs):
        """Constructor

        Args:
            function: function or python formatting string

            **kwargs: reactive models in the function by variable name as keyword
                Changes in these models will trigger recalculation of the function
       """
        if isinstance(function, str):
            format_string = function

            def generator(**kwargs) -> str:
                return format_string.format(
                    **{key: str(variable) for key, variable in kwargs.items()}
                )

            function = generator

        RQText.__init__(self, "")
        RQComputedModel.__init__(self, function, **kwargs)

    def get(self) -> str:
        """Get the computed text"""
        return str(super().get())



from typing import TYPE_CHECKING, Callable, Type, Union, Any

from qtpy.QtWidgets import QBoxLayout, QWidget  # type: ignore

if TYPE_CHECKING:
    from PyQt5.QtWidgets import QBoxLayout, QWidget

from .rqboxlayout import RQBoxLayout
from ..models import RQList


class RQHBoxLayout(RQBoxLayout):
    """Reactive QHBoxLayout"""

    def __init__(self, model: RQList,
                 widget: Union[Type[QWidget], Callable[[Any, RQList], QWidget]], *args, **kwargs):
        """Constructor

        Args:
            model: RQList representing all the items in the layout

            widget: QWidget type that represents each item.
                Can also be a function that accepts the item and list model as arguments,
                and returns the widget instance

            args: arguments to pass to the native pyqt layout

            **kwargs: arguments to pass to the native pyqt layout
        """
        super().__init__(model, widget, QBoxLayout.LeftToRight, *args, **kwargs)



from typing import TYPE_CHECKING, Union

from qtpy.QtWidgets import QPushButton  # type: ignore

if TYPE_CHECKING:
    from PyQt5.QtWidgets import QPushButton

from ..models import RQBool

from .rqwidget import RQWidget


class RQPushButton(RQWidget, QPushButton):
    """Reactive PushButton Widget"""

    def __init__(self,
                 *args,
                 rq_if: Union[RQBool, None] = None,
                 rq_disabled: Union[RQBool, None] = None,
                 **kwargs):
        """Constructor.

        Args:
            model: Model to link the widget to
            *args: arguments to pass to the native pyqt widget
            rq_if: RQBool that controls the visibility
            rq_disabled: RQBool that controls the disabling
            **kwargs: arguments to pass to the native pyqt widget
        """
        RQWidget.__init__(self, model=None, rq_if=rq_if, rq_disabled=rq_disabled)
        QPushButton.__init__(self, *args, **kwargs)
        self.rq_init_widget()



from .rqmodel import RQModel


class RQObject(RQModel):
    """Reactive Object Model

    Represents any object, without a strict type
    """

    def __init__(self, instance: object):
        """Constructor

        Args:
            instance: Initial object
        """
        super(RQObject, self).__init__()
        self._instance = instance

    def get(self) -> object:
        """Get value of the model

        Returns:
            object: object of the model
        """
        return self._instance

    def set(self, instance: object) -> None:
        """Set value of model

        Will propagate the change to the widgets linked to the model

        Args:
            instance: new value of the model
        """
        self._instance = instance
        self.rq_data_changed.emit()

    def __str__(self) -> str:
        """Get value of the model in string format

        Returns:
            str: value of the model converted to string
        """
        return self._instance.__str__()

    def __int__(self) -> int:
        """Get value of the model in int format

        Returns:
            str: value of the model converted to int

        Raises:
            TypeError
        """
        if hasattr(self._instance, "__int__"):
            return self._instance.__int__()  # type: ignore
        else:
            raise TypeError

    def __float__(self) -> float:
        """Get value of the model in float format

        Returns:
            str: value of the model converted to float

        Raises:
            TypeError
        """
        if hasattr(self._instance, "__float__"):
            return self._instance.__float__()  # type: ignore
        else:
            raise TypeError



from .rqbool import RQBool, RQComputedBool
from .rqchoice import RQChoice
from .rqfloat import RQFloat, RQComputedFloat
from .rqint import RQInt, RQComputedInt
from .rqlist import RQList, RQComputedList
from .rqmodel import RQModel, RQComputedModel
from .rqobject import RQObject
from .rqtext import RQText, RQComputedText



from enum import Enum, auto
from typing import TYPE_CHECKING, Union

from qtpy.QtCore import QThread  # type: ignore

if TYPE_CHECKING:
    from PyQt5.QtCore import QThread

from pyreaqtive import RQBool, RQModel


class RQAsync(QThread):
    class AutoTriggers(Enum):
        Start = auto()

    def __init__(self, task, trigger: Union[AutoTriggers, RQModel] = AutoTriggers.Start):
        super(RQAsync, self).__init__()
        self.task = task
        self.trigger = trigger

        self.working = RQBool(False)

        if self.trigger == self.AutoTriggers.Start:
            self.start()
        elif isinstance(trigger, RQModel):
            trigger.rq_data_changed.connect(self.start)

    def run(self) -> None:
        self.working.set(True)
        self.task()
        self.working.set(False)



from typing import TYPE_CHECKING, List, Callable, Type, Union, Any

from qtpy.QtCore import Slot  # type: ignore
from qtpy.QtWidgets import QBoxLayout, QWidget  # type: ignore

if TYPE_CHECKING:
    from PyQt5.QtWidgets import QBoxLayout, QWidget
    from PyQt5.QtCore import pyqtSlot as Slot

from ..models import RQList


class RQBoxLayout(QBoxLayout):
    """Reactive QBoxLayout"""

    def __init__(self, model: RQList,
                 widget: Union[Type[QWidget], Callable[[Any, RQList], QWidget]], *args,
                 **kwargs):
        """Constructor

        Args:
            model: RQList representing all the items in the layout

            widget: QWidget type that represents each item.
                The constructor of the widget must take as arguments: object (Object) and list (RQList).
                Can also be a function that accepts the item and list model as arguments,
                and returns the widget instance

            args: arguments to pass to the native pyqt layout

            **kwargs: arguments to pass to the native pyqt layout
        """
        super().__init__(*args, **kwargs)
        self.model: RQList = model
        """Model linked to the layout"""

        self.widgets: List[QWidget] = []
        """List of current widgets in the layout"""

        self._rq_widget_callback: Callable[[Any, RQList], QWidget]
        """Widget callback. For a new object that is insert on the list, must return the new and appropriate widget"""

        if issubclass(type(widget), QWidget):
            def callback(item: Any, list_model: RQList):
                widget(item, list_model)  # type: ignore

            self._rq_widget_callback = callback
        elif callable(widget):
            self._rq_widget_callback = widget
        else:
            raise TypeError

        if not isinstance(model, RQList):
            raise TypeError
        else:
            self.model.rq_list_insert.connect(self._rq_insert_widget)
            self.model.rq_list_remove.connect(self._rq_remove_widget)

        for index, item in enumerate(self.model):
            self._rq_insert_widget(index)

    @Slot(int)
    def _rq_insert_widget(self, index) -> None:
        """Slot triggered when the model inserts a new item

        Args:
            index: index in the initial of the new instance
        """
        item = self.model[index]
        self.widgets.insert(index, self._rq_widget_callback(item, self.model))
        self.insertWidget(index, self.widgets[index])

    @Slot(int)
    def _rq_remove_widget(self, index) -> None:
        """Slot triggered when the model removes an item

        Args:
            index: index in the list of the item removal
        """
        self.removeWidget(self.widgets[index])
        self.widgets[index].deleteLater()
        self.widgets.pop(index)



from setuptools import setup, find_packages

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()
from pyreaqtive import __version__

setup(
    name='pyreaqtive',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version=__version__,
    packages=find_packages(include=["pyreaqtive", "pyreaqtive.*"]),
    url='https://github.com/polfeliu/pyreaqtive',
    license='Attribution-NoDerivatives 4.0 International (CC BY-ND 4.0)',
    author='Pol Feliu Cuberes',
    author_email='feliupol@gmail.com',
    keywords='pyqt5 framework reactive gui',
    description='PyQt Reactive Framework',
    install_requires=[
        "qtpy>2.0.0"
    ]
)



from typing import TYPE_CHECKING, Any, Callable

from qtpy.QtCore import QObject, Signal, Slot  # type: ignore

if TYPE_CHECKING:
    from PyQt5.QtCore import QObject
    from PyQt5.QtCore import pyqtSignal as Signal
    from PyQt5.QtCore import pyqtSlot as Slot


class RQModel(QObject):
    """RQModel Base Class.

    All pyreaqtive models must inherit from this class, that provides basic get, set method and data changed signals
    """

    rq_read_only = False
    """Indicates if the model cannot be written to. Set() raises error"""

    rq_data_changed = Signal()
    """Data changed signal.

    Widgets that are connected to models can connect slots to this signal.
    The model must emit to this when the state of it changes, to notify the widgets.
    """

    _rq_delete = Signal()
    """Delete signal.

    Signals that the model instance is about to be deleted
    """

    def get(self) -> Any:
        """Method to get the value of the underlying object.

        Must be overridden by model

        Returns:
            Any: value of the object
        """
        raise NotImplementedError

    def set(self, value: Any) -> None:
        """Method to set the value of the underlying object.

        Must be overridden by model

        Args:
            value: New value of the model
        """
        raise NotImplementedError

    def __delete__(self):
        self._rq_delete.emit()


class RQComputedModel:
    """RQComputedModel Base Class.

    All pyreaqtive computed models must inherit from this class.
    Computed models are those that are calculated from others values, by means of a function.
    Changes in the linked models trigger events that updates linked widgets, than in turn calculate the function.
    """

    def __init__(self, function: Callable, **kwargs):
        """Constructor

        Args:
            function: function to calculate the model value from input values

            **kwargs: reactive models in the function by variable name as keyword
                Changes in these models will trigger recalculation of the function
        """
        if not issubclass(type(self), RQModel):
            raise TypeError("RQComputed models must inherit from RQModel by the parent class")

        self.rq_read_only = True
        self.rq_computed_function: Callable = function
        self.rq_computed_variables: dict = kwargs
        for name, model in self.rq_computed_variables.items():
            if isinstance(model, RQModel) or issubclass(type(model), RQModel):
                model.rq_data_changed.connect(self._variable_changed)

        # First calculation
        self._variable_changed()

    @Slot()
    def _variable_changed(self) -> None:
        """Variable changed slot

        Called when some of the models have emitted rq_data_changed.
        Informs connected widgets that the function model has changed.
        Widgets will ask the value again and recalculate it with the new data
        """
        # RQModels have rq_data_changed, asserted on __init__
        self.rq_data_changed.emit()  # type: ignore

    def set(self, value) -> None:
        raise RuntimeError("Computed Models do not allow set()")

    def get(self) -> Any:
        """Get value of the model in the output format of the function

        Returns:
            function result with current model values
        """
        return self.rq_computed_function(
            **{key: variable.get() for key, variable in self.rq_computed_variables.items()}
        )



from typing import Callable

from .rqmodel import RQModel, RQComputedModel


class RQInt(RQModel):
    """Reactive Integer Model
    
    Represents a initial_integer number
    """

    def __init__(self, initial_integer: int):
        """Constructor

        Args:
            initial_integer: Initial value of the model
        """
        super(RQInt, self).__init__()
        self._int: int = int(initial_integer)
        """Model store variable"""

    def get(self) -> int:
        """Get value of the model

        Returns:
            int: value of the model
        """
        return self._int

    def set(self, value: int) -> None:
        """Set value of model

        Will propagate the change to the widgets linked to the model

        Args:
            value: new value of the model
        """
        self._int = int(value)
        self.rq_data_changed.emit()

    def increment(self, delta: int = 1) -> None:
        """Increment integer method

        Args:
            delta: Increment value. Default 1
        """
        self.set(self._int + delta)

    def decrement(self, delta: int = 1):
        """Decrement integer method

        Args:
            delta: Decrement value. Default 1
        """
        self.increment(delta=-delta)

    def __str__(self) -> str:
        """Get value of the model in string format

        Returns:
            str: value of the model converted to string
        """
        return str(self._int)

    def __int__(self) -> int:
        """Get value of the model in integer format

        Returns:
            str: value of the model
        """
        return self.get()

    def __float__(self) -> float:
        """Get value of the model in float format

        Returns:
            str: value of the model converted to float
        """
        return float(self.get())


class RQComputedInt(RQComputedModel, RQInt):
    """Reactive Computed Integer Model"""

    def __init__(self, function: Callable, **kwargs):
        """Constructor

        Args:
            function: function to calculate the model value from input values

            **kwargs: reactive models in the function by variable name as keyword
                Changes in these models will trigger recalculation of the function
       """
        RQInt.__init__(self, 0)
        RQComputedModel.__init__(self, function, **kwargs)

    def get(self) -> int:
        """Get the computed value"""
        return int(super().get())



from .rqcheckbox import RQCheckBox
from .rqcombobox import RQComboBox
from .rqdial import RQDial
from .rqdoublespinbox import RQDoubleSpinBox
from .rqlabel import RQLabel
from .rqlineedit import RQLineEdit
from .rqprogressbar import RQProgressBar
from .rqpushbutton import RQPushButton
from .rqslider import RQSlider
from .rqspinbox import RQSpinBox
from .rqwidget import RQWidget
from .rqwidgetobject import RQWidgetObject



from typing import TYPE_CHECKING, Union

from qtpy.QtCore import Slot  # type: ignore
from qtpy.QtWidgets import QSpinBox  # type: ignore

if TYPE_CHECKING:
    from PyQt5.QtCore import pyqtSlot as Slot
    from PyQt5.QtWidgets import QSpinBox

from .rqwidget import RQWidget
from ..models import RQInt, RQBool, RQObject, RQModel


class RQSpinBox(RQWidget, QSpinBox):
    """Reactive SpinBox Widget"""

    model: Union[RQInt, RQObject]
    """Model linked to the widget"""

    def __init__(self,
                 model: Union[RQInt, int, RQObject],
                 *args,
                 rq_if: Union[RQBool, None] = None,
                 rq_disabled: Union[RQBool, None] = None,
                 wait_for_finish: bool = False,
                 **kwargs
                 ):
        """Constructor.

        Args:
            model: Model to link the widget to
            *args: arguments to pass to the native pyqt widget
            rq_if: RQBool that controls the visibility
            rq_disabled: RQBool that controls the disabling
            wait_for_finish: if true, the model is not updated until Enter is pressed or focus is changed
            **kwargs: arguments to pass to the native pyqt widget
        """
        if isinstance(type(model), RQModel):
            if model.rq_read_only:  # type: ignore
                raise IOError("Cannot connect rqspinbox to a read only model")

        RQWidget.__init__(self, model, rq_if, rq_disabled)
        QSpinBox.__init__(self, *args, **kwargs)
        self.rq_init_widget()

        self._rq_data_changed()
        self.model.rq_data_changed.connect(self._rq_data_changed)

        if wait_for_finish:
            self.editingFinished.connect(self._update_model)
        else:
            self.valueChanged.connect(self._update_model)

    @Slot()
    def _rq_data_changed(self) -> None:
        """Slot triggered when the model changes value.

        Updates value of the spinbox
        """
        if not self._rq_writing:
            self._rq_reading = True
            self.setValue(int(self.model))
            self._rq_reading = False

    _rq_writing = False
    """Flag to signal that this widget is triggering the update and is writing to the model"""

    _rq_reading = False
    """Flag to indicate that the model changed and the widget is reading the model"""

    @Slot()
    def _update_model(self) -> None:
        """Propagates changes to the model"""
        if not self._rq_reading:
            self._rq_writing = True
            self.model.set(self.value())
            self._rq_writing = False
